from dataclasses import dataclass


@dataclass(frozen=True)
class Position:
    x: int
    y: int

    def __add__(self, other) -> "Position":
        return Position(self.x + other.x, self.y + other.y)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
import random

from weiqi.core.figure import Stone
from weiqi.core.position import Position
from weiqi.core.move import Move
from weiqi.core.board import Board

if TYPE_CHECKING:
    from weiqi.core.game import WeiqiGame


class BaseBot(ABC):
    def __init__(self, figure: Stone):
        self._figure = figure

    @property
    def figure(self) -> Stone:
        return self._figure

    @abstractmethod
    def make_move(self, game: "WeiqiGame") -> Move: ...

    def __eq__(self, other) -> bool:
        if not isinstance(other, BaseBot):
            return NotImplemented
        return self.figure == other.figure and type(self) is type(other)


class RandomBot(BaseBot):
    @staticmethod
    def _calc_field_board(matrix: list[list[int]]) -> float:
        """Calculates the field board in percentage from the matrix."""
        count_non_zero = sum(1 for row in matrix for x in row if x != 0)
        total_elements = len(matrix) * len(matrix[0]) if matrix else 0
        return count_non_zero / total_elements if total_elements > 0 else 0.0

    @staticmethod
    def _should_pass_after_opponent_pass(last_move: Move | None) -> bool:
        return (
            last_move is not None
            and last_move.position is None
            and random.random() < 0.4
        )

    @staticmethod
    def _get_random_position(size: int) -> Position:
        x_rand = random.randint(0, size - 1)
        y_rand = random.randint(0, size - 1)
        return Position(x_rand, y_rand)

    def make_move(self, game: "WeiqiGame") -> Move:
        board = game.board
        last_move = game.move_history.last_move

        if self._should_pass_after_opponent_pass(last_move):
            return self._make_pass_move(game)

        if self._should_pass_on_high_occupancy(board):
            return self._make_pass_move(game)

        return self._make_random_valid_move(game, board)

    def _should_pass_on_high_occupancy(self, board: Board) -> bool:
        state_as_matrix = board.state_as_matrix
        fielded_board = self._calc_field_board(state_as_matrix)
        return 0.8 <= fielded_board <= 1.0 and random.random() < 0.4

    def _make_pass_move(self, game: "WeiqiGame") -> Move:
        """Makes a pass move."""
        move = Move(position=None, figure=self.figure)
        game.make_move(self, move)
        return move

    def _make_random_valid_move(
        self, game: "WeiqiGame", board: Board
    ) -> Move:
        """Makes a random valid move."""
        max_attempts = 15
        for attempt in range(max_attempts):
            position = self._get_random_position(board.size)
            move = Move(position=position, figure=self.figure)
            if board.figures[position] is None:
                try:
                    game.make_move(self, move)
                    return move
                except ValueError:
                    continue
        # If no valid move was found, pass.
        return self._make_pass_move(game)



from typing import TYPE_CHECKING

from weiqi.core.figure import Stone
from weiqi.core.position import Position
from weiqi.core.move import Move

if TYPE_CHECKING:
    from weiqi.core.game import WeiqiGame


class Player:
    def __init__(self, figure: Stone):
        self._figure = figure

    @property
    def figure(self) -> Stone:
        return self._figure

    def make_move(
        self, game: "WeiqiGame", position: Position | None
    ) -> None:
        move = Move(position=position, figure=self.figure)
        game.make_move(self, move)

    def resign(self, game: "WeiqiGame") -> None:
        game.resign(self)

    def __eq__(self, other) -> bool:
        if not isinstance(other, Player):
            return NotImplemented
        return self.figure == other.figure



import copy

from weiqi.exceptions.game import GameOverException
from weiqi.core.board import Board
from weiqi.utils.enums import Winner
from weiqi.core.figure import Stone
from weiqi.core.move import MoveHistory, Move
from weiqi.players.player import Player
from weiqi.players.bot import BaseBot
from weiqi.utils.game_status import GameStatus


class WeiqiGame:
    def __init__(
        self,
        board: Board,
        player_black: Player | BaseBot,
        player_white: Player | BaseBot,
        turn: Stone | None = None,
        game_status: GameStatus | None = None,
        move_history: MoveHistory | None = None,
        komi: float | int = 6.5,  # 6.5 is the Japanese and Korean rules.
    ):
        self._board = board
        self._players = [player_black, player_white]
        self._turn = turn or Stone.BLACK
        self._game_status = game_status or GameStatus(False, None)
        self._move_history = move_history or MoveHistory()
        self._komi = komi

        self._validate_players()

    @property
    def board(self) -> Board:
        """Returns a copy of the board."""
        return copy.deepcopy(self._board)

    @property
    def game_status(self) -> GameStatus:
        return self._game_status

    @property
    def players(self) -> list[Player | BaseBot]:
        return self._players

    @property
    def move_history(self) -> MoveHistory:
        return self._move_history

    @property
    def turn(self) -> Stone:
        return self._turn

    @property
    def komi(self) -> float:
        return self._komi

    def _validate_players(self):
        if not all(
            isinstance(player, (Player, BaseBot)) for player in self._players
        ):
            raise ValueError("Invalid player type.")
        if all(isinstance(player, BaseBot) for player in self._players):
            raise ValueError("At least one player must be human.")
        if not all(
            player.figure in (Stone.BLACK, Stone.WHITE)
            for player in self._players
        ):
            raise ValueError("Invalid player color.")
        if len(set(player.figure for player in self._players)) != 2:
            raise ValueError("Players must have different colors.")

    def get_current_player(self) -> Player | BaseBot:
        return next(
            player for player in self._players if player.figure == self._turn
        )

    def resign(self, player: Player):
        if self._game_status.is_over:
            raise GameOverException("Game is already over.")
        if player not in self._players:
            raise ValueError("Invalid player.")
        winner = Winner.WHITE if player.figure == Stone.BLACK else Winner.BLACK
        self._game_status.end_game(winner, None, None)

    def make_move(
        self,
        player: Player | BaseBot,
        move: Move,
    ):
        if self._game_status.is_over:
            raise GameOverException("Game is already over.")

        current_player = self.get_current_player()

        if player != current_player:
            raise ValueError("It's not your turn.")
        if move.figure != player.figure:
            raise ValueError("You can't place a figure of another color.")

        if move.position is not None:
            self._board.place_figure(move)
        else:
            last_move = self._move_history.last_move
            # If the last move was a pass, the game is over.
            if last_move and last_move.position is None:
                score = self._board.score
                black_score = score[Stone.BLACK]
                white_score = score[Stone.WHITE] + self._komi
                if black_score > white_score:
                    winner = Winner.BLACK
                elif white_score > black_score:
                    winner = Winner.WHITE
                else:
                    winner = Winner.DRAW
                self._game_status.end_game(winner, black_score, white_score)

        self._move_history.add_move(move)
        self._next_turn()

    def _next_turn(self):
        self._turn = Stone.BLACK if self._turn == Stone.WHITE else Stone.WHITE



from weiqi.utils.enums import Winner


class GameStatus:
    def __init__(
        self,
        is_over: bool,
        winner: Winner | None,
        black_score: int | None = None,
        white_score: float | int | None = None,
    ):
        self._is_over = is_over
        self._winner = winner

        self._validate_over_game()
        self._validate_scores(black_score, white_score)

        self._black_score = black_score
        self._white_score = white_score

    @property
    def is_over(self) -> bool:
        return self._is_over

    @property
    def winner(self) -> Winner | None:
        return self._winner

    @property
    def black_score(self) -> int | None:
        return self._black_score

    @property
    def white_score(self) -> float | int | None:
        return self._white_score

    def _validate_over_game(self):
        if self._is_over and self._winner is None:
            raise ValueError("Game is over but no winner is set.")

    @staticmethod
    def _validate_scores(
        black_score: int | None, white_score: float | int | None
    ):
        if (black_score is not None) != (white_score is not None):
            raise ValueError("Both scores cannot be set at the same time.")

    def end_game(
        self,
        winner: Winner,
        black_score: int | None,
        white_score: float | int | None,
    ):
        self._is_over = True
        self._winner = winner

        self._validate_scores(black_score, white_score)
        self._black_score = black_score
        self._white_score = white_score



from bisect import insort
from dataclasses import dataclass, field
import datetime

from weiqi.core.position import Position
from weiqi.core.figure import Stone


@dataclass(frozen=True)
class Move:
    position: Position | None
    figure: Stone
    timestamp: datetime.datetime = field(init=False)

    def __post_init__(self):
        utc_tz = datetime.timezone.utc
        object.__setattr__(self, "timestamp", datetime.datetime.now(utc_tz))


class MoveHistory:
    def __init__(self, history: list[Move] | None = None):
        self._history = history or []
        self._history.sort(key=lambda m: m.timestamp)

    def add_move(self, move: Move):
        insort(self._history, move, key=lambda m: m.timestamp)

    @property
    def last_move(self) -> Move | None:
        if self._history:
            return self._history[-1]
        return None

    def get_all_moves(self) -> list[Move]:
        return self._history

    def __iter__(self):
        return iter(self._history)

    def __len__(self):
        return len(self._history)

    def __getitem__(self, item):
        return self._history[item]



from itertools import product

from weiqi.core.group import Group
from weiqi.core.position import Position
from weiqi.core.figure import Stone
from weiqi.core.move import Move


class Board:
    """Class for the board of the Weiqi game."""

    def __init__(
        self,
        figures: dict[Position, Stone | None] | str | list[list[int]],
        white_captured: int = 0,
        black_captured: int = 0,
    ):
        if isinstance(figures, str):
            self._figures = self._from_string(figures)
        elif isinstance(figures, list):
            self._figures = self._from_matrix(figures)
        else:
            self._figures = figures

        self._size = int(len(self._figures) ** 0.5)
        self._white_captured = white_captured
        self._black_captured = black_captured

        if not self._validate_available_size():
            raise ValueError("Not available size.")
        if not self._validate_positions():
            raise ValueError("Invalid positions.")
        if not self._is_square_board():
            raise ValueError("Board must be square.")
        if not self._validate_figures():
            raise ValueError("Invalid figures.")

        for group in self._find_groups_without_liberties():
            self.__remove_group(group)

    @property
    def figures(self) -> dict[Position, Stone | None]:
        return self._figures

    @property
    def size(self) -> int:
        return self._size

    @property
    def white_captured(self) -> int:
        """
        Number of white stones captured by black player (stones).
        """
        return self._white_captured

    @property
    def black_captured(self) -> int:
        """
        Number of black stones captured by white player (stones).
        """
        return self._black_captured

    def _is_square_board(self) -> bool:
        unique_x = len(set(position.x for position in self._figures.keys()))
        unique_y = len(set(position.y for position in self._figures.keys()))
        return unique_x == unique_y and unique_x == self._size

    def _validate_positions(self) -> bool:
        return all(
            position.x < self._size and position.y < self._size
            for position in self._figures.keys()
        )

    def _validate_figures(self) -> bool:
        return all(
            isinstance(stone, (Stone, type(None)))
            for stone in self._figures.values()
        )

    def _validate_available_size(self) -> bool:
        valid_sizes = {5, 6, 7, 8, 9, 11, 13, 15, 17, 19}
        return self.size in valid_sizes

    def position_in_bounds(self, position: Position) -> bool:
        return 0 <= position.x < self._size and 0 <= position.y < self._size

    def _find_groups_without_liberties(self) -> set[Group]:
        return {
            self._group_at_position(position)
            for position in self._get_not_empty_positions()
            if not self._group_at_position(position).liberties
        }

    def _get_not_empty_positions(self) -> list[Position]:
        return [
            position
            for position, stone in self._figures.items()
            if stone is not None
        ]

    def _get_neighbors(self, position: Position) -> list[Position]:
        return [
            position + delta
            for delta in [
                Position(0, 1),
                Position(0, -1),
                Position(1, 0),
                Position(-1, 0),
            ]
            if self.position_in_bounds(position + delta)
        ]

    def __remove_group(self, group: Group):
        count = len(group.positions)

        if group.figure == Stone.BLACK:
            self._black_captured += count
        elif group.figure == Stone.WHITE:
            self._white_captured += count

        for position in group.positions:
            self._figures[position] = None

    def _group_at_position(self, position: Position) -> Group:
        figure = self._figures.get(position, None)
        if figure is None:
            raise ValueError("Position is empty.")

        def bfs(queue: list[Position], visited: set[Position], group: Group):
            while queue:
                pos = queue.pop(0)
                if pos in visited:
                    continue
                visited.add(pos)

                if self._figures.get(pos) == figure:
                    group.positions.add(pos)
                    queue.extend(
                        neighbor
                        for neighbor in self._get_neighbors(pos)
                        if neighbor not in visited
                    )
                elif self._figures.get(pos) is None:
                    group.liberties.add(pos)
            return group

        group = Group(positions=set(), liberties=set(), figure=figure)
        return bfs([position], set(), group)

    def find_territories(self) -> dict[Stone | None, set[Position]]:
        visited: set[Position] = set()
        white_territory: set[Position] = set()
        black_territory: set[Position] = set()
        neutral_territory: set[Position] = set()

        def dfs(
            position: Position,
            visited: set[Position],
            territory: set[Position],
            colors: set[Stone],
        ):
            stack = [position]
            while stack:
                current_position = stack.pop()
                if current_position in visited:
                    continue
                visited.add(current_position)
                territory.add(current_position)

                for neighbor in self._get_neighbors(current_position):
                    if neighbor in visited:
                        continue

                    neighbor_stone = self._figures.get(neighbor)
                    if neighbor_stone is None:
                        stack.append(neighbor)
                    else:
                        colors.add(neighbor_stone)

        for i in range(self.size):
            for j in range(self.size):
                position = Position(i, j)
                if (
                    self._figures[position] is None
                    and position not in neutral_territory
                ):
                    colors: set[Stone] = set()
                    territory: set[Position] = set()
                    dfs(position, visited, territory, colors)

                    if Stone.BLACK in colors and Stone.WHITE not in colors:
                        black_territory.update(territory)
                    elif Stone.WHITE in colors and Stone.BLACK not in colors:
                        white_territory.update(territory)
                    else:
                        neutral_territory.update(territory)

        return {
            Stone.BLACK: black_territory,
            Stone.WHITE: white_territory,
            None: neutral_territory,
        }

    @property
    def score(self) -> dict[Stone, int]:
        territories = self.find_territories()
        black_score = len(territories[Stone.BLACK])
        white_score = len(territories[Stone.WHITE])

        max_figures = self.size**2
        expected_score = max_figures - 1

        # If one figure in board, score for white and black is 0,0 respectively
        if expected_score == black_score or expected_score == white_score:
            return {
                Stone.BLACK: 0,
                Stone.WHITE: 0,
            }

        return {
            Stone.BLACK: black_score + self.white_captured,
            Stone.WHITE: white_score + self.black_captured,
        }

    @staticmethod
    def generate_empty_board(size: int) -> "Board":
        figures: dict[Position, Stone | None] = {
            Position(x, y): None for x, y in product(range(size), range(size))
        }
        return Board(figures)

    @property
    def state_as_matrix(self) -> list[list[int]]:
        """
        Converts the board state to a matrix representation.

        -1 - white stone
        0 - empty intersection
        1 - black stone

        Returns:
            list[list[int]]: The board state as a matrix.
        """
        state = [[0] * self.size for _ in range(self.size)]
        for position, stone in self._figures.items():
            state[position.y][position.x] = (
                1
                if stone == Stone.BLACK
                else -1 if stone == Stone.WHITE else 0
            )
        return state

    @property
    def state_as_string(self) -> str:
        """
        Converts the board state to a string representation.

        W - white stone
        B - black stone
        . - empty intersection

        Returns:
            str: The board state as a string.
        """
        symbols = {Stone.BLACK: "B", Stone.WHITE: "W", None: "."}
        return "/".join(
            "".join(
                symbols.get(self._figures.get(Position(x, y)), ".")
                for x in range(self.size)
            )
            for y in range(self.size)
        )

    @staticmethod
    def _from_matrix(matrix: list[list[int]]) -> dict[Position, Stone | None]:
        return {
            Position(x, y): (
                Stone.BLACK
                if cell == 1
                else Stone.WHITE if cell == -1 else None
            )
            for y, row in enumerate(matrix)
            for x, cell in enumerate(row)
        }

    @staticmethod
    def _from_string(string: str) -> dict[Position, Stone | None]:
        return Board._from_matrix(
            [
                [
                    1 if cell == "B" else -1 if cell == "W" else 0
                    for cell in row
                ]
                for row in string.split("/")
            ]
        )

    def place_figure(self, move: Move) -> None:
        if move.position is None:
            raise ValueError("Position is required.")
        if not self.position_in_bounds(move.position):
            raise ValueError("Position out of bounds.")
        if self._figures.get(move.position) is not None:
            raise ValueError("Intersection occupied by existing stone.")
        white_captured = self.white_captured
        black_captured = self.black_captured
        figures = self._figures.copy()

        self._figures[move.position] = move.figure

        try:
            neighboring_enemy_groups = [
                self._group_at_position(neighbor)
                for neighbor in self._get_neighbors(move.position)
                if self._figures.get(neighbor) not in {None, move.figure}
            ]

            for group in neighboring_enemy_groups:
                if not group.liberties:
                    self.__remove_group(group)

            new_group = self._group_at_position(move.position)
        except ValueError as e:
            self._figures = figures
            self._white_captured = white_captured
            self._black_captured = black_captured
            raise e

        if not new_group.liberties:
            self._figures = figures
            self._white_captured = white_captured
            self._black_captured = black_captured
            raise ValueError("New group has zero liberties (suicide)")



from weiqi.core.figure import Stone
from weiqi.core.position import Position


class Group:
    def __init__(
        self, positions: set[Position], liberties: set[Position], figure: Stone
    ):
        self.positions = positions
        self.liberties = liberties
        self.figure = figure

    def __hash__(self):
        return hash(
            (frozenset(self.positions), frozenset(self.liberties), self.figure)
        )

    def __eq__(self, other):
        if not isinstance(other, Group):
            return False
        return (
            frozenset(self.positions) == frozenset(other.positions)
            and frozenset(self.liberties) == frozenset(other.liberties)
            and self.figure == other.figure
        )



from weiqi.core.game import WeiqiGame
from weiqi.core.board import Board
from weiqi.core.figure import Stone
from weiqi.core.position import Position
from weiqi.core.move import Move, MoveHistory
from weiqi.players.player import Player
from weiqi.players.bot import BaseBot, RandomBot
from weiqi.utils.game_status import GameStatus
from weiqi.utils.enums import Winner


__all__ = [
    "WeiqiGame",
    "Board",
    "Stone",
    "Position",
    "Move",
    "MoveHistory",
    "Player",
    "BaseBot",
    "RandomBot",
    "GameStatus",
    "Winner",
]
